#include <stddef.h>
#include <math.h>

#include "matrix.h"

/* matrices */
void            mult_matrix(left, right, dest)
matrix          left, right, dest;
{
   int             i, j, k;
   matrix          temp;

   zero_matrix(temp);
   for (i = 0; i < 4; i++)
   {
      for (j = 0; j < 4; j++)
      {
         for (k = 0; k < 4; k++)
            temp[i][j] += left[i][k] * right[k][j];
      }
   }
   copy_matrix(dest, temp);
}

void            transpose(dest, srce)
matrix          dest, srce;
{
   register int    i, j;

   for (i = 0; i < 4; i++)
   {
      dest[i][0] = srce[0][i];
      dest[i][1] = srce[1][i];
      dest[i][2] = srce[2][i];
      dest[i][3] = srce[3][i];
   }
}

void            zero_matrix(dest)
matrix          dest;
{
   vector_zero(dest[0]);
   vector_zero(dest[1]);
   vector_zero(dest[2]);
   vector_zero(dest[3]);
}

void            id_matrix(dest)
matrix          dest;
{
   register int    i, j;

   for (i = 0; i < 4; i++)
   {
      for (j = 0; j < 4; j++)
         dest[i][j] = ((i == j) ? 1 : 0);
   }
}

void            copy_matrix(dest, srce)
matrix          dest, srce;
{
   vector_copy(dest[0], srce[0]);
   vector_copy(dest[1], srce[1]);
   vector_copy(dest[2], srce[2]);
   vector_copy(dest[3], srce[3]);
}

void            vector_transform(d, s, m)
vector          d, s;
matrix          m;

/* performs matrix multiplication on a vector */
/* x   0   1   2   3    srce=  0 1 2 3        */
/* 0  00  01  02  03    dest= 00+.. 30        */
/* 1  10  11  12  13          01+.. 31        */
/* 2  20  21  22  23          02+.. 32        */
/* 3  30  31  32  33          03+.. 33        */
{
   register int    i;
   vector          temp;

   vector_zero(temp);
   for (i = 0; i < 4; i++)
   {
      temp[0] += m[i][0] * s[i];
      temp[1] += m[i][1] * s[i];
      temp[2] += m[i][2] * s[i];
      temp[3] += m[i][3] * s[i];
   }
   vector_copy(d, temp);
}

void            translate(dest, v)
matrix          dest;
vector          v;
{
   id_matrix(dest);
   dest[3][0] = v[0];
   dest[3][1] = v[1];
   dest[3][2] = v[2];
}

void            scale(dest, v)
matrix          dest;
vector          v;
{
   id_matrix(dest);
   dest[0][0] = v[0];
   dest[1][1] = v[1];
   dest[2][2] = v[2];
}

void            rotate_x(dest, angle)
matrix          dest;
double          angle;
{
   id_matrix(dest);
   dest[1][1] = cos(angle);
   dest[1][2] = sin(angle);
   dest[2][1] = -dest[1][2];
   dest[2][2] = dest[1][1];
}

void            rotate_y(dest, angle)
matrix          dest;
double          angle;
{
   id_matrix(dest);
   dest[0][0] = cos(angle);
   dest[0][2] = -sin(angle);
   dest[2][0] = -dest[0][2];
   dest[2][2] = dest[0][0];
}

void            rotate_z(dest, angle)
matrix          dest;
double          angle;
{
   id_matrix(dest);
   dest[0][0] = cos(angle);
   dest[0][1] = sin(angle);
   dest[1][0] = -dest[0][1];
   dest[1][1] = dest[0][0];
}

void            print_matrix(m)
matrix          m;
{
   int             i, j;

   for (i = 0; i < 4; i++)
   {
      for (j = 0; j < 4; j++)
         printf("%f ", m[i][j]);
      printf("\n");
   }
   printf("\n");
}

void            print_vector(v)
vector          v;
{
   printf("<%f %f %f %f>\n", v[0], v[1], v[2], v[3]);
}

void            print_point(p)
point          *p;
{
   printf("%f %f %f\n", p->x, p->y, p->z);
}

void            points_copy(dest, srce, n)
point           dest[], srce[];
int             n;
{
   register int    i;

   for (i = 0; i < n; i++)
   {
      dest[i].x = srce[i].x;
      dest[i].y = srce[i].y;
      dest[i].z = srce[i].z;
   }
}

void            unit_vector(dest, srce)
vector          dest, srce;
{
   double          R;

   R = sqrt(SQR(srce[0]) + SQR(srce[1]) + SQR(srce[2]));
   double2vector(dest, (srce[0] / R), (srce[1] / R), (srce[2] / R));
}

void            unit_line_vector(dest, head, tail)
vector          dest;
point          *head, *tail;
{
   double          x, y, z, R;

   x = head->x - tail->x;
   y = head->y - tail->y;
   z = head->z - tail->z;
   R = sqrt(SQR(x) + SQR(y) + SQR(z));
   double2vector(dest, (x / R), (y / R), (z / R));
}

void            normalized_cross_product(dest, v1, v2)
vector          dest, v1, v2;
{
   double          x, y, z, inv_R;

   x = (v1[1] * v2[2]) - (v1[2] * v2[1]);
   y = (v1[2] * v2[0]) - (v1[0] * v2[2]);
   z = (v1[0] * v2[1]) - (v1[1] * v2[0]);
   inv_R = 1.0 / sqrt(SQR(x) + SQR(y) + SQR(z));
   double2vector(dest, (x * inv_R), (y * inv_R), (z * inv_R));
}

/* the following can be defined as macros ************/

#ifndef USE_MACROS
void            vector_copy(dest, srce)
vector          dest, srce;
{
   register int    i;

   for (i = 0; i < 4; i++)
      dest[i] = srce[i];
}

void            point_copy(dest, srce)
point          *dest, *srce;
{
   dest->x = srce->x;
   dest->y = srce->y;
   dest->z = srce->z;
}

void            double2point(dest, v1, v2, v3)
point          *dest;
double          v1, v2, v3;
{
   dest->x = v1;
   dest->y = v2;
   dest->z = v3;
}

void            point2double(srce, v1, v2, v3)
point          *srce;
double         *v1, *v2, *v3;
{
   *v1 = srce->x;
   *v2 = srce->y;
   *v3 = srce->z;
}

void            double2vector(dest, v1, v2, v3)
vector          dest;
double          v1, v2, v3;
{
   dest[0] = v1;
   dest[1] = v2;
   dest[2] = v3;
   dest[3] = 1.0;
}

void            vector2double(srce, v1, v2, v3)
vector          srce;
double         *v1, *v2, *v3;
{
   *v1 = srce[0];
   *v2 = srce[1];
   *v3 = srce[2];
}

void            point2vector(dest, srce)
vector          dest;
point          *srce;
{
   dest[0] = srce->x;
   dest[1] = srce->y;
   dest[2] = srce->z;
   dest[3] = 1.0;
}

void            vector2point(dest, srce)
point          *dest;
vector          srce;
{
   dest->x = srce[0];
   dest->y = srce[1];
   dest->z = srce[2];
}

void            homogenize(dest)
vector          dest;
{
   dest[0] /= dest[3];
   dest[1] /= dest[3];
   dest[2] /= dest[3];
   dest[3] = 1.0;
}

void            vector_init(dest)
vector          dest;
{
   dest[0] = 0.0;
   dest[1] = 0.0;
   dest[2] = 0.0;
   dest[3] = 1.0;
}

void            vector_zero(dest)
vector          dest;
{
   dest[0] = 0.0;
   dest[1] = 0.0;
   dest[2] = 0.0;
   dest[3] = 0.0;
}

void            line_vector(dest, head, tail)
vector          dest;
point          *head, *tail;
{
   dest[0] = head->x - tail->x;
   dest[1] = head->y - tail->y;
   dest[2] = head->z - tail->z;
   dest[3] = 1.0;
}

void            dot_product(dest, v1, v2)
double         *dest;
vector          v1, v2;
{
   *dest = (v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]);
}

#endif

/***************************************/

void homogenize_vector_transform(vector v, matrix Matrix)
{
   register double *Matrix_row_0 = Matrix[0];
   register double *Matrix_row_1 = Matrix[1];
   register double *Matrix_row_2 = Matrix[2];
   register double *Matrix_row_3 = Matrix[3];

   double x_hat = v[X];
   double y_hat = v[Y];
   double z_hat = v[Z];
   double h_hat = v[H];
   double h;

   h = 1.0 / (Matrix_row_3[0] * x_hat + Matrix_row_3[1] * y_hat
              + Matrix_row_3[2] * z_hat + Matrix_row_3[3] * h_hat);

   v[X] = (Matrix_row_0[0] * x_hat + Matrix_row_0[1] * y_hat
         + Matrix_row_0[2] * z_hat + Matrix_row_0[3] * h_hat) * h;
   v[Y] = (Matrix_row_1[0] * x_hat + Matrix_row_1[1] * y_hat
         + Matrix_row_1[2] * z_hat + Matrix_row_1[3] * h_hat) * h;
   v[Z] = (Matrix_row_2[0] * x_hat + Matrix_row_2[1] * y_hat
         + Matrix_row_2[2] * z_hat + Matrix_row_2[3] * h_hat) * h;
   v[H] = 1.0;
}

void vector_transform1(vector Vector, matrix Matrix)
{
   double *Matrix_row_0 = Matrix[0];
   double *Matrix_row_1 = Matrix[1];
   double *Matrix_row_2 = Matrix[2];
   double *Matrix_row_3 = Matrix[3];

   double x_hat = Vector[0];
   double y_hat = Vector[1];
   double z_hat = Vector[2];
   double h_hat = Vector[3];

   Vector[0] = Matrix_row_0[0] * x_hat + Matrix_row_0[1] * y_hat
               + Matrix_row_0[2] * z_hat + Matrix_row_0[3] * h_hat;
   Vector[1] = Matrix_row_1[0] * x_hat + Matrix_row_1[1] * y_hat
               + Matrix_row_1[2] * z_hat + Matrix_row_1[3] * h_hat;
   Vector[2] = Matrix_row_2[0] * x_hat + Matrix_row_2[1] * y_hat
               + Matrix_row_2[2] * z_hat + Matrix_row_2[3] * h_hat;
   Vector[3] = Matrix_row_3[0] * x_hat + Matrix_row_3[1] * y_hat
               + Matrix_row_3[2] * z_hat + Matrix_row_3[3] * h_hat;

}
